/****************************************************
*  File: GA.cpp
*  Purpose: Contains the Genetic Algorithm Logic
*
*  Date:    Author:    Notes:
*  050505   morelli    Initial rev
*
*
*
****************************************************/




#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "GA.h"

#define ACC 1000


static int debug = 0;

/********************************************************/
GA::GA()
{
  srand(time(NULL));
  crossoverRate = 0;
  mutationRate = 0;
  population = 0;


}
/********************************************************/

GA::~GA()
{
 //delete citizens, and their vectors...


}
/********************************************************/
int GA::ProcessConfigFile()
{
  FILE * fp;
  char type[100];
  double value;
  fp = fopen ("./config.cfg", "ro");
  if (!fp)
    return (-1);
  while (fscanf(fp, "%s", type) != EOF)
  {
    fscanf(fp, "%lf", &value);
    if (!(strcmp(type, "XOVER")))
    {
      crossoverRate = value;
    }
    if (!(strcmp(type, "MUTAT")))
    {
      mutationRate = value;
    }
    if (!(strcmp(type, "POPUL")))
    {
      population = (int)value;
    }
    if (!(strcmp(type, "VECSZ")))
    {
      vecSize = (int)value;
    }
    if (!(strcmp(type, "GENER")))
    {
      generations = (int)value;
    }
    if (!(strcmp(type, "XPLY")))
    {
      xPly = (int)value;
    }
    if (!(strcmp(type, "OPLY")))
    {
      oPly = (int)value;
    }
  }
  // Should do some checking here to know that we read everything....
  fclose (fp);
  return 0;
}
/********************************************************/
int GA::run()
{
  printf("Starting GA\n");
  if (ProcessConfigFile())
  {
    printf("Error processing config file\n");
    return (-1);
  }  
  InitVectors();
  RunGA();

  return 0;


}
/********************************************************/
int GA::InitVectors()
{
  if (debug)
  {
    printf("InitVectors Called: %d\n", vecSize);
  }
  xCitizens = (Citizen *) malloc (population * (sizeof(Citizen)));
  xCitizensTemp = (Citizen *) malloc (population * (sizeof(Citizen)));

  oCitizens = (Citizen *) malloc (population * (sizeof(Citizen)));
  oCitizensTemp = (Citizen *) malloc (population * (sizeof(Citizen)));
  for (int x = 0; x< population; x++)
  {
    xCitizens[x].chrom = (int *) malloc (vecSize * (sizeof(int)));
    xCitizensTemp[x].chrom = (int *) malloc (vecSize * (sizeof(int)));
    memset(xCitizens[x].chrom, 0, vecSize);
    memset(xCitizensTemp[x].chrom, 0, vecSize);
    xCitizens[x].value = 0;
    xCitizensTemp[x].value = 0;

    oCitizens[x].chrom = (int *) malloc (vecSize * (sizeof(int)));
    oCitizensTemp[x].chrom = (int *) malloc (vecSize * (sizeof(int)));
    memset(oCitizens[x].chrom, 0, vecSize);
    memset(oCitizensTemp[x].chrom, 0, vecSize);
    oCitizens[x].value = 0;
    oCitizensTemp[x].value = 0;

    for (int y = 0; y < vecSize; y++)
    {
      xCitizensTemp[x].chrom[y] = xCitizens[x].chrom[y] = getrandom(0,1);
      oCitizensTemp[x].chrom[y] = oCitizens[x].chrom[y] = getrandom(0,1);
    }

  }
  for (int x = 0; x< 5; x++)
  {
    topX[x] = (int *) malloc (vecSize * sizeof(int));
    topO[x] = (int *) malloc (vecSize * sizeof(int));
    for (int y = 0; y< vecSize; y++)
    {
      //Initialize them to something for the first pass..
      //This is as random as anything...

      topX[x][y] = xCitizens[x].chrom[y];
      topO[x][y] = oCitizens[x].chrom[y];
    }

  }
  return 0;
}

/********************************************************/
int GA::RunGA()
{
  int * mates; 
  mates = (int *) malloc (population * sizeof(int));
  memset (mates, 0, population);
  for ( int x = 0; x < generations; x++)
  {
//    if (debug)
    {
      printf("Generation: %d\n", x); 
      fflush(stdout);
    }
    for (int y=0; y < population; y++)
    { 
      xCitizensTemp[y].value = xCitizens[y].value = 
                                Eval(xCitizens[y].chrom, XTURN);
      oCitizensTemp[y].value = oCitizens[y].value = 
                                Eval(oCitizens[y].chrom, OTURN);
    }
    double * xValues = (double *) malloc (population * sizeof(double));
    double * oValues = (double *) malloc (population * sizeof(double));
    for (int x = 0; x< population; x++)
    {
      xValues[x] = StaticEval(xCitizens[x].chrom);  
      oValues[x] = StaticEval(xCitizens[x].chrom);  
    }
//    DumpStats(xValues);
//    DumpStats(oValues);
    free (xValues);
    free (oValues);

    FindMates(xCitizens, mates);
    for (int i = 0; i < population ; i++)
    {
      if (getrandom(0,ACC) < crossoverRate * ACC)
        XOverMutate(xCitizens[i].chrom, xCitizensTemp[mates[i]].chrom);
    }
    FindMates(oCitizens, mates);
    for (int i = 0; i < population ; i++)
    {
      if (getrandom(0,ACC) < crossoverRate * ACC)
        XOverMutate(oCitizens[i].chrom, oCitizensTemp[mates[i]].chrom);
    }


    for (int x = 0; x< population; x++)
    {
      memcpy(xCitizensTemp[x].chrom, xCitizens[x].chrom, 
                                         vecSize * sizeof (int));
      memcpy(oCitizensTemp[x].chrom, oCitizens[x].chrom, 
                                         vecSize * sizeof (int));
    }

  }
  return 0;
}

/********************************************************/
double GA::Eval(int * _vec, int _turn)
{
  double value = 0.0;
  double weights[95];  //Hard Coding is a bad thing...
  double thresholds[23];

  double oppWeights[95];
  double oppThresholds[23];


  Net * net;
  Net * oppNet;
  TicTacToe * ticTacToe;
  for (int x = 0; x< 5; x++)
  {
    Decode (_vec, weights, thresholds);
    
    if (_turn == XTURN)
    {
      Decode (topO[x], oppWeights, oppThresholds);
    }
    else
    {
      Decode (topX[x], oppWeights, oppThresholds);
    }

    net = new Net(weights, thresholds);
    oppNet = new Net(oppWeights, oppThresholds);
    ticTacToe = new TicTacToe(net, oppNet);
    ticTacToe->SetXPly(xPly);
    ticTacToe->SetOPly(oPly);

    value += ticTacToe->PlayGame();
    if (debug)
    {
      printf("Returned from playgame\n");
      fflush(stdout);
    }
    if (_turn == OTURN)
    {
      value *=-1;   //Negative is good...
    }
    value += 240; //lowest possible score is -240... 
    delete ticTacToe;
    delete net;
    delete oppNet;
  }
  return value;
}
/********************************************************/
double GA::StaticEval(int * _vec)
{
  double value = 0.0;
  double weights[95];  //Hard Coding is a bad thing...
  double thresholds[23];

  Net * net;
  TicTacToe * ticTacToe;
  Decode (_vec, weights, thresholds);
    
  net = new Net(weights, thresholds);
  ticTacToe = new TicTacToe(net);
  ticTacToe->SetXPly(xPly);
  ticTacToe->SetOPly(oPly);

  value += ticTacToe->PlayGame();
  if (debug)
  {
    printf("Returned from playgame\n");
    fflush(stdout);
  }
  value *=-1;   //Negative is good...
  value += 240; //lowest possible score is -240... 

  delete ticTacToe;
  delete net;

  return value;
}
/********************************************************/

void GA::Decode (int * _vec, double * _weights, double * _thresholds)
{
  int tempval;
  int vecPointer = 0;
  for (int x = 0; x<95; x++)
  {
    tempval = 0;
    for (int y = 0; y < 5; y++)
    {
      if (_vec[vecPointer])
        tempval |= 1<<y;
      vecPointer++;
    }
    _weights[x] = -8 + double(tempval) / 2;
  }

  for (int x = 0; x<23; x++)
  {
    tempval = 0;
    for (int y = 0; y < 5; y++)
    {
      if (_vec[vecPointer])
        tempval |= 1<<y;
      vecPointer++;
    }
    _thresholds[x] = -8 + double(tempval) / 2;
  }

  if (debug)
  {
    for (int x = 0; x< 23; x++)
    {
      printf("%lf\t%lf\n", _weights[x], _thresholds[x]);

    }
    printf("Total bits %d\n", vecPointer);
  }
 
}

/********************************************************/
void GA::FindMates(Citizen * _cit, int * _mates)
{
  double totalFitness = 0;
  int  * ranges = (int *)malloc (population * sizeof(int));

  for (int x = 0; x< population; x++)
  {
    totalFitness += _cit[x].value;
    ranges[x] = (int)totalFitness;
  }
  for (int x = 0; x< population; x++)
  {
//Potential issue here person can mate with himself
    int location = getrandom(0,(int)totalFitness);
    _mates[x] = population;
    for (int y = 0; y< population ; y++)
    {
      if (location <= ranges[y])
      {
        _mates[x] = y;
        break;
      }
    }
  }
}



/********************************************************/
void GA::XOverMutate(int * _vecA, int * _vecB)
{
  int x;
  int xOverPoint = getrandom(0, vecSize-1);
  for (x = 0; x < vecSize; x++)
  {
    if (x > xOverPoint)
      _vecA[x] = _vecB[x];

    if (!(getrandom(0,int(1/mutationRate))))
    {
      if (_vecA[x] == 0)
        _vecA[x] = 1;
      else
        _vecA[x] = 0;
    }
  }
}


/********************************************************/
void GA::DumpStats(double * _values)
{
  double totalFitness = 0;
  double maxFitness = -100000;
  double minFitness = 100000;
  for (int x = 0; x< population; x++)
  {
    totalFitness += _values[x];
    if (minFitness > _values[x])
      minFitness = _values[x];
    if (maxFitness < _values[x])
      maxFitness = _values[x];
  }
  double avgFitness = totalFitness / population;
  printf("Avg: %lf, Max: %lf, Min: %lf\n", avgFitness, maxFitness, minFitness);
  fflush(stdout);
}
/********************************************************/

