/****************************************************
*  File: Net.cpp
*  Purpose: Logic for constructing a neural net for Tic Tac Toe
*
*  Date:    Author:    Notes:
*  050505   morelli    Initial rev
*
*
*
****************************************************/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Net.h"
#include "Node.h"

int debug = 0;

/***********************************************************/
Net::Net(double * _weights, double * _thresholds)
{

  memcpy(weights, _weights, 95*sizeof(double));
  memcpy(thresholds, _thresholds, 23*sizeof(double));

  GenerateNet();



}

/***********************************************************/
Net::~Net()
{

  free (levelNodes);

  for (int x = 0; x < numInputs; x++)
  {
    delete (inputNodes[x]);
  }
  for (int x = 0; x < totalNodes; x++)
  {
    delete (nodeMap[x]);
  }
}
/***********************************************************/
void Net::GenerateNet()
{
  numInputs = 9;
  numLevels = 3;
  levelNodes = 0;
  totalNodes = 0;


  inputNodes = (Node **) malloc (numInputs * sizeof (Node*));
  for (int x = 0; x < numInputs; x++)
  {
    inputNodes[x] = new Node();
  }

  levelNodes = (int *)malloc(numLevels * sizeof(int));

  levelNodes[0] = 18;
  levelNodes[1] = 5;
  levelNodes[2] = 1;
  totalNodes = 18+5+1;

  nodeMap = (Node **) malloc (totalNodes * sizeof (Node*));
  for (int x = 0; x < totalNodes; x++)
  {
    nodeMap[x] = new Node();
  }
  for (int x = 0; x<numInputs; x++)
  {
    int temp = 0;
    int totalOutputs = 0;
    
    for (int y = 0; y<18; y++)   
    {
      inputNodes[x]->outPuts[totalOutputs] = nodeMap[y];
      totalOutputs ++;
    }
  }

  int weightPtr = 0;
  for (int x = 0; x<18; x++)
  {
    int temp = 0;
    int totalOutputs = 0;
    for (int y = 18; y<23; y++) 
    {
      nodeMap[x]->outPuts[totalOutputs] = nodeMap[y];
      nodeMap[x]->weight[totalOutputs] = weights[weightPtr];
      weightPtr++;
      totalOutputs ++;
    }
  }
  for (int x = 18; x<23; x++)
  {
    int temp = 0;
    int totalOutputs = 0;
    nodeMap[x]->outPuts[totalOutputs] = nodeMap[23];
    nodeMap[x]->weight[totalOutputs] = weights[weightPtr];
    weightPtr++;
    totalOutputs ++;
  }
  
  for (int x = 0; x<totalNodes; x++)
  {
    nodeMap[x]->threshold = thresholds[x];
  }

}
/***********************************************************/
double Net::Eval()
{
  for (int x = 0; x < numInputs; x++)
  {
    double tempIn = (double)inputValues[x];
    int y = 0;
    while (inputNodes[x]->outPuts[y])
    {
      inputNodes[x]->outPuts[y]->input += tempIn;
      y++;
    }
  }

  for (int x = 0; x < totalNodes; x++)
  {
    double tempIn = 0.0;
    int y = 0;
    while (nodeMap[x]->outPuts[y])
    {
      if (nodeMap[x]->input > nodeMap[x]->threshold)
        nodeMap[x]->outPuts[y]->input += nodeMap[x]->weight[y];
      y++;
    }
  }

/*
  for (int x =0; x<totalNodes; x++)
  {
    if (nodeMap[x]->outPuts[0] == 0)
      printf("Node %d Value: %lf\n", x, nodeMap[x]->input);

  }
*/
  if (debug)
    printf("BoardVal: %lf\n", nodeMap[23]->input);
  return (nodeMap[23]->input);

}
/***********************************************************/
void Net::SetInputValues(int * _inputs)
{

  for (int x = 0; x< 9; x++)
  {
    inputValues[x] = _inputs[x];
  }
}
/***********************************************************/
