/****************************************************
*  File: TicTacToe.cpp
*  Purpose: Plays a game of tic tac toe
*
*  Date:    Author:    Notes:
*  050505   morelli    Initial rev
*
*
*
****************************************************/




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "TicTacToe.h"

static int debug = 0;
int temp =0;


/************************************************************/
TicTacToe::TicTacToe(Net * _net, Net * _oppNet)
{
  pNet = _net;
  pOppNet = _oppNet;
  board = new Board(0);
  board->xLocation = 0;
  board->oLocation = 0;
}

/************************************************************/
TicTacToe::TicTacToe(Net * _net)
{
  pNet = _net;
  pOppNet = 0;
  board = new Board(0);
  board->xLocation = 0;
  board->oLocation = 0;
}

/************************************************************/
TicTacToe::~TicTacToe()
{
  if (board)
    delete board;


}
/************************************************************/
double TicTacToe::PlayGame()
{
  int winner = NOWINS;
  int turn = OTURN;
  int move;
  double retVal;

  Board * currentBoard;
  int place = getrandom(0,8);
  board->xLocation |= 1<< place;

  while (winner == NOWINS)
  {
    GenerateChildren(board, turn, 0, oPly);
    move = FindMove(board, oPly, OTURN);
    if (move == -1)
      assert(0);
    currentBoard = new Board(0);
    currentBoard->xLocation = board->children[move]->xLocation;
    currentBoard->oLocation = board->children[move]->oLocation;
    delete (board);
    board = 0;

    winner = CheckForWinner(currentBoard);
    if (winner != NOWINS)
    {
//      DumpBoard(currentBoard);
      retVal = StaticEval(currentBoard, true);
      break;
      //Calc final board and return...
    }
    GenerateChildren(currentBoard, XTURN, 0, xPly);
    move = FindMove(currentBoard, xPly, XTURN); 

    board = new Board(0);
    board->xLocation = currentBoard->children[move]->xLocation;
    board->oLocation = currentBoard->children[move]->oLocation;
    delete (currentBoard);
    currentBoard = 0;

    winner = CheckForWinner(board);
    if (winner != NOWINS)
    {
      //Calc final board and return...
      retVal = StaticEval(board, true);
      break;

    }


    if (debug)
    {
      Pause();
      DumpBoard(board);
    }
  }
  if (debug)
  {
    printf("Play Game Returning: %lf\n", retVal);
    fflush(stdout);
    DumpBoard(board);
    Pause();
  }
  return retVal;
}
/************************************************************/
int TicTacToe::GenerateChildren(Board * _parent, int _turn, 
                                int _depth, int _maxDepth)
{
  unsigned long xLocation = _parent->xLocation;
  unsigned long oLocation = _parent->oLocation;
  int totalChildren = 0;
  int tokenPlaced = 0;
  int nextTurn = 0;
  int depth = _depth;
  if (_depth < _maxDepth)
  {
    depth++;
    for (int x = 0; x<9; x++)
    {
      if (!(xLocation & 1<<x) && !(oLocation & 1<<x))
      {
        if (debug)
        {
          printf("Placing token @ %d Turn: %d\n", x, _turn);
        }
        tokenPlaced = 1;
        _parent->children[totalChildren] = new Board (_parent);
        if (_turn == XTURN)
        {
          _parent->children[totalChildren]->oLocation = oLocation;
          _parent->children[totalChildren]->xLocation = xLocation | 1<<x;
        }
        else
        {
          _parent->children[totalChildren]->xLocation = xLocation;
          _parent->children[totalChildren]->oLocation = oLocation | 1<<x;
        }
        totalChildren++;
        temp++;
        if (_turn == XTURN)
          nextTurn = OTURN;
        else
          nextTurn = XTURN;

        GenerateChildren(_parent->children[totalChildren-1], nextTurn, depth,
                          _maxDepth);
      }
    }
  }
  return 0;
}
/************************************************************/
void TicTacToe::SetXPly(int _ply)
{
  xPly = _ply;
}

/************************************************************/
void TicTacToe::SetOPly(int _ply)
{
  oPly = _ply;
}

/************************************************************/
void TicTacToe::DumpBoard(Board * _board)
{
  PrintBoard(_board);
  for (int x = 0; x<9; x++)
  {
    if (_board->children[x])
    {
      if (debug)
      {
        printf("DumpBoard - Child Found: %d\n", x);
      }
      DumpBoard(_board->children[x]);
    }
  }
}
/************************************************************/
void TicTacToe::PrintBoard(Board * _board)
{
  printf("=====\n");
  for (int x = 0; x<9; x++)
  {
    if (_board->xLocation & 1<<x)
      printf("X");
    else if (_board->oLocation & 1<<x)
      printf("O");
    else
      printf(" ");
    if (x==2 || x==5 || x==8)
      printf("\n");
  }
  printf("=====\n");
  fflush(stdout);

}
/************************************************************/
int TicTacToe::FindMove(Board * _board, int _ply, int _turn)
{
  if (_turn == XTURN)
    SetLeafValues(_board, 0, _ply, _turn);
  else
    SetLeafValues(_board, 0, _ply, _turn);

  for (int x = _ply; x>0; x--)
  {
    if (x%2 == 0)
    {
      SetParentValue (_board, 0, x-1, MAX_VALUE);
      InvestigateNodes(_board, 0, x, MINIMIZE);
    }
    else
    {
      SetParentValue (_board, 0, x-1, MIN_VALUE);
      InvestigateNodes(_board, 0, x, MAXIMIZE);
    }

  }
  for (int x = 0; x< MAXNODES; x++)
  {
    if (_board->children[x])
    {
      if (_board->children[x]->score == _board->score)
      {
        return x;
      }
    }
  }
  return -1;
}
/************************************************************/
void TicTacToe::SetLeafValues (Board * _board, int _depth, int _maxDepth, 
                               int _turn)
{
  int inputs[9];
  if (_depth < _maxDepth)
  {
    _depth++;
    for (int x = 0; x< MAXNODES; x++)
    {
      if (_board->children[x])
        SetLeafValues(_board->children[x], _depth, _maxDepth, _turn);
    }
  }
  else
  {
    if (_turn == OTURN)
    {
      for (int x = 0; x<9; x++)
      {
        if (_board->xLocation & 1<<x)
          inputs[x] = 1;
        else if (_board->oLocation & 1<<x)
          inputs[x] = -1;
        else
          inputs[x] = 0;
      }
      pNet->SetInputValues(inputs);
      _board->score = pNet->Eval();
    }
    else if (pOppNet)
    {
      for (int x = 0; x<9; x++)
      {
        if (_board->xLocation & 1<<x)
          inputs[x] = 1;
        else if (_board->oLocation & 1<<x)
          inputs[x] = -1;
        else
          inputs[x] = 0;
      }
      pOppNet->SetInputValues(inputs);
      _board->score = pOppNet->Eval();
    }
    else
      _board->score = StaticEval(_board);
    if (debug)
    {
      DumpBoard(_board);
      printf("Score: %lf\n", _board->score);
    }
  }
}
/************************************************************/
void TicTacToe::SetParentValue (Board * _board, int _depth, 
                                int _maxDepth, int _value)
{

  if (_depth < _maxDepth)
  {
    _depth++;
    for (int x = 0; x< MAXNODES; x++)
    {
      if (_board->children[x])
        SetParentValue(_board->children[x], _depth, _maxDepth, _value);
    }
  }
  else
  {
    _board->score = (double)_value;
  }
}
/************************************************************/
void TicTacToe::InvestigateNodes(Board * _board, int _depth, 
                                 int _maxDepth, int _type)
{
  if (_depth < _maxDepth)
  {
    _depth++;
    for (int x = 0; x< MAXNODES; x++)
    {
      if (_board->children[x])
        InvestigateNodes(_board->children[x], _depth, _maxDepth, _type);
    }
  }
  else
  {
    if (_type == MINIMIZE)
    {
      if (_board->parent->score > _board->score)
      {
        _board->parent->score = _board->score;
      }
    }
    else
    {
      if (_board->parent->score < _board->score)
      {
        _board->parent->score = _board->score;
      }
    }

  }
}


/************************************************************/
void TicTacToe::Pause()
{
  int unused;
  printf("App is paused\n");
  scanf("%d", &unused);
}
/************************************************************/
double TicTacToe::StaticEval(Board * _board, bool _final)
{
  unsigned long xLoc = _board->xLocation;
  unsigned long oLoc = _board->oLocation;
  double boardVal = 0;
  int totalWins = 0;
  int totalAlmostWins = 0;
  int totalBlocks = 0; 
  int totalAlmostLosses = 0;
  int totalLosses = 0;

  for (int x = 0; x<TOTAL_WIN_PATTERNS; x++)
  {
    if ((xLoc & WINPATTERNS[x]) == WINPATTERNS[x])
    {
      if (debug)
      {
        printf("Found Win Pattern for x @ %d, %ld, %ld, %ld\n", x, xLoc, 
                              WINPATTERNS[x], xLoc & WINPATTERNS[x]);
        fflush(stdout);
      }
      totalWins++;
    }
  }
  for (int x = 0; x<TOTAL_WIN_PATTERNS; x++)
  {
    int winLocs[3];
    int totalLocsFound = 0;
    for (int y = 0; y< 9; y++)
    {
      if (WINPATTERNS[x] & 1<<y)
      {
        winLocs[totalLocsFound] = y;
        totalLocsFound ++;
        if (totalLocsFound ==3)
          break;
      }
    }
    if (totalLocsFound != 3)
    {
      printf("MAJOR PROBLEM IN EVALBOARD!!\n");
      fflush(stdout);
      assert(0);
    }
// These give us the chance to win on the next turn
    if (( xLoc & 1<< winLocs[0]) &&
        ( xLoc & 1<< winLocs[1]) &&
        !(oLoc & 1<< winLocs[2]))
      totalAlmostWins++;
    if (( xLoc & 1<< winLocs[0]) &&
        ( xLoc & 1<< winLocs[2]) &&
        !(oLoc & 1<< winLocs[1]))
      totalAlmostWins++;
    if (( xLoc & 1<< winLocs[1]) &&
        ( xLoc & 1<< winLocs[2]) &&
        !(oLoc & 1<< winLocs[0]))
      totalAlmostWins++;

//These give our opponent the chance to win on the next turn
    if (( oLoc & 1<< winLocs[0]) &&
        ( oLoc & 1<< winLocs[1]) &&
        !(xLoc & 1<< winLocs[2]))
      totalAlmostLosses++;
    if (( oLoc & 1<< winLocs[0]) &&
        ( oLoc & 1<< winLocs[2]) &&
        !(xLoc & 1<< winLocs[1]))
      totalAlmostLosses++;
    if (( oLoc & 1<< winLocs[1]) &&
        ( oLoc & 1<< winLocs[2]) &&
        !(xLoc & 1<< winLocs[0]))
      totalAlmostLosses++;


    //Blocks should be taken into account when looking for potential moves
    //when evaluating final board score, they are not used as they are
    //the opposite of an almost win for the opponent, they cancel each
    //other out...
    if (!_final)
    {
      //These will be considered blocks...
      if (( oLoc & 1<< winLocs[0]) &&
          ( oLoc & 1<< winLocs[1]) &&
          ( xLoc & 1<< winLocs[2]))
        totalBlocks++;
      if (( oLoc & 1<< winLocs[0]) &&
          ( oLoc & 1<< winLocs[2]) &&
          ( xLoc & 1<< winLocs[1]))
        totalBlocks++;
      if (( oLoc & 1<< winLocs[1]) &&
          ( oLoc & 1<< winLocs[2]) &&
          ( xLoc & 1<< winLocs[0]))
        totalBlocks++;
    }
//These are losses...
    if (( oLoc & 1<< winLocs[0]) &&
        ( oLoc & 1<< winLocs[1]) &&
        ( oLoc & 1<< winLocs[2]))
      totalLosses++;
    if (( oLoc & 1<< winLocs[0]) &&
        ( oLoc & 1<< winLocs[2]) &&
        ( oLoc & 1<< winLocs[1]))
      totalLosses++;
    if (( oLoc & 1<< winLocs[1]) &&
        ( oLoc & 1<< winLocs[2]) &&
        ( oLoc & 1<< winLocs[0]))
      totalLosses++;
  }
  if (debug)
  {
    printf("%d %d %d %d %d\n", totalWins, totalAlmostWins, totalBlocks, 
                          totalAlmostLosses, totalLosses);


  }
  boardVal = WWIN * totalWins + WAWIN*totalAlmostWins+WBLK*totalBlocks + 
             WALOSS * totalAlmostLosses + WLOSS*totalLosses;



  return boardVal;
}
/************************************************************/
int TicTacToe::CheckForWinner(Board * _board)
{
  unsigned long xLocation = _board->xLocation;
  unsigned long oLocation = _board->oLocation;


  for (int x = 0; x< TOTAL_WIN_PATTERNS; x++)
  {
    if ((xLocation & WINPATTERNS[x]) == WINPATTERNS[x])
    {
      return XWINS;
    }
    if ((oLocation & WINPATTERNS[x]) == WINPATTERNS[x])
    {
      return OWINS;
    }
  }
  if ((xLocation | oLocation) == FULL_CARD)
  {
    return TIE;
  }
  return NOWINS;
}

/************************************************************/
